@@ -102,4 +102,4 @@ echo ""
 echo "Success! kafkacat is now built"
 echo ""
 
-./kafkacat -V
+./kafkacat -h
